@@ -17,9 +17,11 @@ import base64
 import mock
 import copy
 import os
+import json
 
 import botocore
 import botocore.session
+from botocore.compat import OrderedDict
 from botocore.exceptions import ParamValidationError, MD5UnavailableError
 from botocore.exceptions import AliasConflictParameterError
 from botocore.awsrequest import AWSRequest
@@ -29,7 +31,7 @@ from botocore.docs.bcdoc.restdoc import DocumentStructure
 from botocore.docs.params import RequestParamsDocumenter
 from botocore.docs.example import RequestExampleDocumenter
 from botocore.hooks import HierarchicalEmitter
-from botocore.model import OperationModel, ServiceModel
+from botocore.model import OperationModel, ServiceModel, ServiceId
 from botocore.model import DenormalizedStructureBuilder
 from botocore.signers import RequestSigner
 from botocore.credentials import Credentials
@@ -137,7 +139,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'ec2', 'us-east-1', 'ec2', 'v4', credentials, event_emitter)
+            ServiceId('ec2'), 'us-east-1', 'ec2', 'v4',
+            credentials, event_emitter)
         handlers.inject_presigned_url_ec2(
             params, request_signer, operation_model)
         self.assertEqual(params['body']['PresignedUrl'], 'https://foo')
@@ -154,7 +157,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'ec2', 'us-east-1', 'ec2', 'v4', credentials, event_emitter)
+            ServiceId('ec2'), 'us-east-1', 'ec2', 'v4', credentials,
+            event_emitter)
         handlers.inject_presigned_url_ec2(
             params, request_signer, operation_model)
         self.assertEqual(params['body']['PresignedUrl'], 'https://foo')
@@ -167,7 +171,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'rds', 'us-east-1', 'rds', 'v4', credentials, event_emitter)
+            ServiceId('rds'), 'us-east-1', 'rds', 'v4', credentials,
+            event_emitter)
         handlers.inject_presigned_url_rds(
             params, request_signer, operation_model)
         self.assertEqual(params['body']['PreSignedUrl'], 'https://foo')
@@ -184,7 +189,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'rds', 'us-east-1', 'rds', 'v4', credentials, event_emitter)
+            ServiceId('rds'), 'us-east-1', 'rds', 'v4', credentials,
+            event_emitter)
         handlers.inject_presigned_url_rds(
             params, request_signer, operation_model)
         self.assertEqual(params['body']['PreSignedUrl'], 'https://foo')
@@ -196,7 +202,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'ec2', 'us-east-1', 'ec2', 'v4', credentials, event_emitter)
+            ServiceId('ec2'), 'us-east-1', 'ec2', 'v4', credentials,
+            event_emitter)
         request_dict = {}
         params = {'SourceRegion': 'us-west-2'}
         request_dict['body'] = params
@@ -248,7 +255,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'ec2', actual_region, 'ec2', 'v4', credentials, event_emitter)
+            ServiceId('ec2'), actual_region, 'ec2', 'v4', credentials,
+            event_emitter)
         request_dict = {}
         params = {
             'SourceRegion': 'us-west-2',
@@ -277,7 +285,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'rds', 'us-east-1', 'rds', 'v4', credentials, event_emitter)
+            ServiceId('rds'), 'us-east-1', 'rds', 'v4', credentials,
+            event_emitter)
         request_dict = {}
         params = {'SourceRegion': 'us-west-2'}
         request_dict['body'] = params
@@ -303,7 +312,9 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'rds', 'us-east-1', 'rds', 'v4', credentials, event_emitter)
+            ServiceId('rds'), 'us-east-1', 'rds', 'v4', credentials,
+            event_emitter
+        )
         request_dict = {}
         params = {'SourceRegion': 'us-west-2'}
         request_dict['body'] = params
@@ -326,7 +337,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'rds', 'us-east-1', 'rds', 'v4', credentials, event_emitter)
+            ServiceId('rds'), 'us-east-1', 'rds', 'v4', credentials,
+            event_emitter)
         request_dict = {}
         params = {'SourceRegion': 'us-west-2', 'PreSignedUrl': 'https://foo'}
         request_dict['body'] = params
@@ -349,7 +361,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'rds', 'us-east-1', 'rds', 'v4', credentials, event_emitter)
+            ServiceId('rds'), 'us-east-1', 'rds', 'v4', credentials,
+            event_emitter)
         request_dict = {}
         params = {'SourceRegion': 'us-west-2'}
         request_dict['body'] = params
@@ -373,7 +386,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'rds', 'us-east-1', 'rds', 'v4', credentials, event_emitter)
+            ServiceId('rds'), 'us-east-1', 'rds', 'v4', credentials,
+            event_emitter)
         handlers.inject_presigned_url_rds(
             params=params,
             request_signer=request_signer,
@@ -387,7 +401,8 @@ class TestHandlers(BaseSessionTest):
         credentials = Credentials('key', 'secret')
         event_emitter = HierarchicalEmitter()
         request_signer = RequestSigner(
-            'rds', 'us-east-1', 'rds', 'v4', credentials, event_emitter)
+            ServiceId('rds'), 'us-east-1', 'rds', 'v4', credentials,
+            event_emitter)
         request_dict = {}
         params = {'SourceRegion': 'us-west-2'}
         request_dict['body'] = params
@@ -534,7 +549,7 @@ class TestHandlers(BaseSessionTest):
                   'UserData': b64_user_data}
         self.assertEqual(params, result)
 
-    def test_register_retry_for_handlers_with_no_endpoint_prefix(self):
+    def test_register_retry_for_handlers_with_no_metadata(self):
         no_endpoint_prefix = {'metadata': {}}
         session = mock.Mock()
         handlers.register_retries_for_service(service_data=no_endpoint_prefix,
@@ -544,7 +559,10 @@ class TestHandlers(BaseSessionTest):
 
     def test_register_retry_handlers(self):
         service_data = {
-            'metadata': {'endpointPrefix': 'foo'},
+            'metadata': {
+                'endpointPrefix': 'foo',
+                'serviceId': 'foo',
+            },
         }
         session = mock.Mock()
         loader = mock.Mock()
@@ -575,6 +593,24 @@ class TestHandlers(BaseSessionTest):
         # an error response.
         self.assertEqual(original, handler_input)
 
+    def test_does_decode_template_body_in_order(self):
+        expected_ordering = OrderedDict([
+            ('TemplateVersion', 1.0),
+            ('APropertyOfSomeKind', 'a value'),
+            ('list', [1, 2, 3]),
+            ('nested', OrderedDict([('key', 'value'),
+                                    ('foo', 'bar')]))
+        ])
+        template_string = json.dumps(expected_ordering)
+        parsed_response = {'TemplateBody': template_string}
+
+        handlers.json_decode_template_body(parsed=parsed_response)
+        result = parsed_response['TemplateBody']
+
+        self.assertTrue(isinstance(result, OrderedDict))
+        for element, expected_element in zip(result, expected_ordering):
+            self.assertEqual(element, expected_element)
+
     def test_decode_json_policy(self):
         parsed = {
             'Document': '{"foo": "foobarbaz"}',
@@ -710,8 +746,6 @@ class TestHandlers(BaseSessionTest):
         handlers.switch_host_with_param(request, 'PredictEndpoint')
         self.assertEqual(request.url, new_endpoint)
 
-
-
     def test_invalid_char_in_bucket_raises_exception(self):
         params = {
             'Bucket': 'bad/bucket/name',
@@ -839,6 +873,70 @@ class TestHandlers(BaseSessionTest):
         handlers.decode_list_object(parsed, context=context)
         self.assertEqual(parsed['Delimiter'], u'\xe7\xf6s% asd\x08 c')
 
+    def test_decode_list_objects_v2(self):
+        parsed = {
+            'Contents': [{'Key': "%C3%A7%C3%B6s%25asd%08"}],
+            'EncodingType': 'url',
+        }
+        context = {'encoding_type_auto_set': True}
+        handlers.decode_list_object_v2(parsed, context=context)
+        self.assertEqual(parsed['Contents'][0]['Key'], u'\xe7\xf6s%asd\x08')
+
+    def test_decode_list_objects_v2_does_not_decode_without_context(self):
+        parsed = {
+            'Contents': [{'Key': "%C3%A7%C3%B6s%25asd"}],
+            'EncodingType': 'url',
+        }
+        handlers.decode_list_object_v2(parsed, context={})
+        self.assertEqual(parsed['Contents'][0]['Key'], u'%C3%A7%C3%B6s%25asd')
+
+    def test_decode_list_objects_v2_with_delimiter(self):
+        parsed = {
+            'Delimiter': "%C3%A7%C3%B6s%25%20asd%08+c",
+            'EncodingType': 'url',
+        }
+        context = {'encoding_type_auto_set': True}
+        handlers.decode_list_object_v2(parsed, context=context)
+        self.assertEqual(parsed['Delimiter'], u'\xe7\xf6s% asd\x08 c')
+
+    def test_decode_list_objects_v2_with_prefix(self):
+        parsed = {
+            'Prefix': "%C3%A7%C3%B6s%25%20asd%08+c",
+            'EncodingType': 'url',
+        }
+        context = {'encoding_type_auto_set': True}
+        handlers.decode_list_object_v2(parsed, context=context)
+        self.assertEqual(parsed['Prefix'], u'\xe7\xf6s% asd\x08 c')
+
+    def test_decode_list_objects_v2_does_not_decode_continuationtoken(self):
+        parsed = {
+            'ContinuationToken': "%C3%A7%C3%B6s%25%20asd%08+c",
+            'EncodingType': 'url',
+        }
+        context = {'encoding_type_auto_set': True}
+        handlers.decode_list_object_v2(parsed, context=context)
+        self.assertEqual(
+            parsed['ContinuationToken'], u"%C3%A7%C3%B6s%25%20asd%08+c")
+
+    def test_decode_list_objects_v2_with_startafter(self):
+        parsed = {
+            'StartAfter': "%C3%A7%C3%B6s%25%20asd%08+c",
+            'EncodingType': 'url',
+        }
+        context = {'encoding_type_auto_set': True}
+        handlers.decode_list_object_v2(parsed, context=context)
+        self.assertEqual(parsed['StartAfter'], u'\xe7\xf6s% asd\x08 c')
+
+    def test_decode_list_objects_v2_with_common_prefixes(self):
+        parsed = {
+            'CommonPrefixes': [{'Prefix': "%C3%A7%C3%B6s%25%20asd%08+c"}],
+            'EncodingType': 'url',
+        }
+        context = {'encoding_type_auto_set': True}
+        handlers.decode_list_object_v2(parsed, context=context)
+        self.assertEqual(parsed['CommonPrefixes'][0]['Prefix'],
+                         u'\xe7\xf6s% asd\x08 c')
+
     def test_get_bucket_location_optional(self):
         # This handler should no-op if another hook (i.e. stubber) has already
         # filled in response
@@ -846,6 +944,50 @@ class TestHandlers(BaseSessionTest):
         handlers.parse_get_bucket_location(response, None),
         self.assertEqual(response["LocationConstraint"], "eu-west-1")
 
+    def test_set_operation_specific_signer_no_auth_type(self):
+        signing_name = 'myservice'
+        context = {'auth_type': None}
+        response = handlers.set_operation_specific_signer(
+            context=context, signing_name=signing_name)
+        self.assertIsNone(response)
+
+    def test_set_operation_specific_signer_unsigned(self):
+        signing_name = 'myservice'
+        context = {'auth_type': 'none'}
+        response = handlers.set_operation_specific_signer(
+            context=context, signing_name=signing_name)
+        self.assertEqual(response, botocore.UNSIGNED)
+
+    def test_set_operation_specific_signer_v4(self):
+        signing_name = 'myservice'
+        context = {'auth_type': 'v4'}
+        response = handlers.set_operation_specific_signer(
+            context=context, signing_name=signing_name)
+        self.assertEqual(response, 'v4')
+
+    def test_set_operation_specific_signer_s3v4(self):
+        signing_name = 's3'
+        context = {'auth_type': 'v4'}
+        response = handlers.set_operation_specific_signer(
+            context=context, signing_name=signing_name)
+        self.assertEqual(response, 's3v4')
+
+    def test_set_operation_specific_signer_v4_unsinged_payload(self):
+        signing_name = 'myservice'
+        context = {'auth_type': 'v4-unsigned-body'}
+        response = handlers.set_operation_specific_signer(
+            context=context, signing_name=signing_name)
+        self.assertEqual(response, 'v4')
+        self.assertEqual(context.get('payload_signing_enabled'), False)
+
+    def test_set_operation_specific_signer_s3v4_unsigned_payload(self):
+        signing_name = 's3'
+        context = {'auth_type': 'v4-unsigned-body'}
+        response = handlers.set_operation_specific_signer(
+            context=context, signing_name=signing_name)
+        self.assertEqual(response, 's3v4')
+        self.assertEqual(context.get('payload_signing_enabled'), False)
+
 
 class TestConvertStringBodyToFileLikeObject(BaseSessionTest):
     def assert_converts_to_file_like_object_with_bytes(self, body, body_bytes):
@@ -997,7 +1139,7 @@ class TestAddMD5(BaseMD5Test):
     def test_adds_md5_when_v4(self):
         credentials = Credentials('key', 'secret')
         request_signer = RequestSigner(
-            's3', 'us-east-1', 's3', 'v4', credentials, mock.Mock())
+            ServiceId('s3'), 'us-east-1', 's3', 'v4', credentials, mock.Mock())
         request_dict = {'body': b'bar',
                         'url': 'https://s3.us-east-1.amazonaws.com',
                         'method': 'PUT',
@@ -1010,7 +1152,8 @@ class TestAddMD5(BaseMD5Test):
     def test_adds_md5_when_s3v4(self):
         credentials = Credentials('key', 'secret')
         request_signer = RequestSigner(
-            's3', 'us-east-1', 's3', 's3v4', credentials, mock.Mock())
+            ServiceId('s3'), 'us-east-1', 's3', 's3v4', credentials,
+            mock.Mock())
         request_dict = {'body': b'bar',
                         'url': 'https://s3.us-east-1.amazonaws.com',
                         'method': 'PUT',
@@ -1039,7 +1182,7 @@ class TestAddMD5(BaseMD5Test):
     def test_add_md5_raises_error_when_md5_unavailable(self):
         credentials = Credentials('key', 'secret')
         request_signer = RequestSigner(
-            's3', 'us-east-1', 's3', 's3', credentials, mock.Mock())
+            ServiceId('s3'), 'us-east-1', 's3', 's3', credentials, mock.Mock())
         request_dict = {'body': b'bar',
                         'url': 'https://s3.us-east-1.amazonaws.com',
                         'method': 'PUT',
@@ -1053,7 +1196,7 @@ class TestAddMD5(BaseMD5Test):
     def test_adds_md5_when_s3v2(self):
         credentials = Credentials('key', 'secret')
         request_signer = RequestSigner(
-            's3', 'us-east-1', 's3', 's3', credentials, mock.Mock())
+            ServiceId('s3'), 'us-east-1', 's3', 's3', credentials, mock.Mock())
         request_dict = {'body': b'bar',
                         'url': 'https://s3.us-east-1.amazonaws.com',
                         'method': 'PUT',
@@ -1177,3 +1320,46 @@ class TestParameterAlias(unittest.TestCase):
         contents = self.sample_section.flush_structure().decode('utf-8')
         self.assertIn(self.alias_name + '=',  contents)
         self.assertNotIn(self.original_name + '=', contents)
+
+
+class TestCommandAlias(unittest.TestCase):
+    def test_command_alias(self):
+        alias = handlers.ClientMethodAlias('foo')
+        client = mock.Mock()
+        client.foo.return_value = 'bar'
+
+        response = alias(client=client)()
+        self.assertEqual(response, 'bar')
+
+
+class TestPrependToHost(unittest.TestCase):
+    def setUp(self):
+        self.hoister = handlers.HeaderToHostHoister('test-header')
+
+    def _prepend_to_host(self, url, prepend_string):
+        params = {
+            'headers': {
+                'test-header': prepend_string,
+            },
+            'url': url,
+        }
+        self.hoister.hoist(params=params)
+        return params['url']
+
+    def test_does_prepend_to_host(self):
+        prepended = self._prepend_to_host('https://bar.example.com/', 'foo')
+        self.assertEqual(prepended, 'https://foo.bar.example.com/')
+
+    def test_does_prepend_to_host_with_http(self):
+        prepended = self._prepend_to_host('http://bar.example.com/', 'foo')
+        self.assertEqual(prepended, 'http://foo.bar.example.com/')
+
+    def test_does_prepend_to_host_with_path(self):
+        prepended = self._prepend_to_host(
+            'https://bar.example.com/path', 'foo')
+        self.assertEqual(prepended, 'https://foo.bar.example.com/path')
+
+    def test_does_prepend_to_host_with_more_components(self):
+        prepended = self._prepend_to_host(
+            'https://bar.baz.example.com/path', 'foo')
+        self.assertEqual(prepended, 'https://foo.bar.baz.example.com/path')
